// --- [ TWI_Master_Slave.c ] -------------------------------------------------
//       20.07.13		tab = 3
//       auf der Grundlage der AppNote AVR 311 / 315 von Atmel
//       fr TWI-Master und TWI-Slave zusammengefgt,
//       aber ohne Multi-Master Fhigkeit !
//       Der Slave dient nur zum Debuggen bzw. als "Terminal" fr die Datenausgabe
//       eines TWI-Masters.
//
//       Master- und Slavefunktionen verfgen ber getrennte Ringbuffer von je 256 Byte
//
//
// ----------------------------------------------------------------------------

#include <avr/interrupt.h>
#include <avr/io.h>
#include <inttypes.h>
#include "TWI_Master_Slave.h"
#include "global.h"


union TWI_statusReg TWI_statusReg = {0};     // TWI_statusReg is defined in TWI_Slave.h

// --- lokale Funktionen
uint8_t TWI_Transceiver_Busy(void);
uint8_t TWI_Slave_Busy(void);
uint8_t TWI_Master_Busy(void);

// --- lokale Variablen
volatile uint8_t twi_head;
volatile uint8_t twi_tail;
volatile uint8_t repeated_start = FALSE;

uint8_t TWI_state    = TWI_NO_STATE;	         // State byte. Default set to TWI_NO_STATE.
volatile uint8_t TWI_Mode;
volatile uint8_t slave_last_state = RECEIVE;    // Voreinstellung: der Slave hat Daten gesendet

// ----------------------------------------------------------------------------
// Grundlegende Initialisierung fr Master und Slave Modus
// Den Unterschied zwischen Master und Slave macht lediglich das Bit TWEA
// (damit antwortet der Slave mit einem ACK, sobald er addressiert wird)
// Fr den Master Modus mssen TWIE / TWINT gesetzt werden
// Fr den Slave Modus  muss zustzlich TWEA gesetzt werden
void TWI_Initialise( void )
{
	TWI_Mode = 0;
   TWBR = TWI_TWBR;                             // Set bit rate register (Baudrate). Defined in header file.
// TWSR = TWI_TWPS;                             // Not used. Driver presumes prescaler to be 00.
	TWAR = TWI_SLA_ADR;                          // Set own TWI slave address.
	TWDR = 0xFF;                                 // Default content = SDA released.
	TWCR = (1<<TWEN)|                          	// Enable TWI-interface and release TWI pins.
          (0<<TWIE)|(0<<TWINT)|                 // Disable TWI Interupt.
          (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|      // Do not ACK on any requests, yet.
          (0<<TWWC);
}

// ---------------------------------------------------------------------------
// Setzt das Flag repeated_start auf True,
// das nachfolgende TWI-Write wird ohne STOP abgeschlossen,
// so dass ein Repeated Start beim anschlieenden TWI-Read entsteht.
// Das Flag wird automatisch am Ende des TWI-Write gelscht.
void TWI_Set_Repeated_Start(void)
{
   repeated_start = TRUE;
}

// ----------------------------------------------------------------------------
// liefert .T., wenn der Master ODER der Slave aktiv ist
// liefert .F., wenn keiner von beiden aktiv
uint8_t TWI_Transceiver_Busy( void )
{
	uint8_t i;
	i  = ((TWI_Mode & (1<<MASTER | 1<<ACTIV)) == (1<<MASTER | 1<<ACTIV));
	i += ((TWI_Mode & (1<<SLAVE  | 1<<ACTIV)) == (1<<SLAVE  | 1<<ACTIV));

	return (i);
}


// ----------------------------------------------------------------------------
// liefert .T. wenn der Master aktiv ist

uint8_t TWI_MA_Busy(void)
{
	return ((TWI_Mode & (1<<MASTER | 1<<ACTIV)) == (1<<MASTER | 1<<ACTIV));
}

// ----------------------------------------------------------------------------
// liefert .T. wenn der Slave aktiv ist

uint8_t TWI_SLA_Busy(void)
{
	return ((TWI_Mode & (1<<SLAVE | 1<<ACTIV)) == (1<<SLAVE | 1<<ACTIV));
}


// ----------------------------------------------------------------------------
uint8_t TWI_Get_Err(void)
{													 // Wait until TWI has completed the transmission.
	while ((TWI_MA_Busy()) || TWI_SLA_Busy());
	if (TWI_state == TWI_NO_STATE) return(0);
	return(TWI_state);
}

/****************************************************************************
Call this function to send a prepared message. The first byte must contain the slave address and the
read/write bit. Consecutive bytes contain the data to be sent, or empty locations for data to be read
from the slave. Also include how many bytes that should be sent/read including the address byte.
The function will hold execution (loop) until the TWI_ISR has completed with the previous operation,
then initialize the next operation and return.
****************************************************************************/
/*
Die Daten stehem im globalen Array rs_buf[255].
Der Start des zu sendenden Bereiches ist tail.
Beim Write werden solange Bytes mit Index tail++ gesendet, bis tail == head
Beim Read werden Bytes mit Index tail++ gelesen, bis tail = head

Die Arbeit erfolgt interruptgesteuert:
*/

void TWI_MA_Transmit_Data( uint8_t tail, uint8_t head, uint8_t * buf )
// Ist zustndig fr das TWI-Read und Write
{

   while ( TWI_Transceiver_Busy() );           	// Wait until Master is ready for next transmission.

   TWI_statusReg.all = 0;
   TWI_state         = TWI_NO_STATE ;
   twi_buf = buf;

	twi_tail = tail;
	twi_head = head;

	TWI_Mode = (1<<MASTER | 1<< ACTIV);
   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|      // Initiate a START condition.
          (0<<TWWC);
}



/****************************************************************************
Call this function to start the Transceiver without specifing new transmission data.
Usefull for restarting a transmission, or just starting the transceiver for reception.
The driver will reuse the data previously put in the transceiver buffers.
The function will hold execution (loop) until the TWI_ISR has completed with the
previous operation, then initialize the next operation and return.
****************************************************************************/
void TWI_SLA_Start_Transceiver( void )
{
	while ( TWI_Transceiver_Busy() );             // Wait until TWI is ready for next transmission.
	TWI_statusReg.all = 0;
	TWI_state         = TWI_NO_STATE ;
	TWI_Mode = (1<<SLAVE);                        // der Slave ist bereit, aber noch nicht aktiv
	TWCR = (1<<TWEN)|                             // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                  // Enable TWI Interupt and clear the flag.
          (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|       // Prepare to ACK next time the Slave is addressed.
          (0<<TWWC);                             //
}

/****************************************************************************
Liefert zurck, ob der Slave Daten empfangen hat.
Solange eine SLA-bertragung nicht abgeschlossen ist, wird FALSE zurckgeliefert.
Wenn vom Slave Daten abgeholt wurden, dann ist slave_last_state == TRANSMIT,
dann wird ebenfalls False zurckgegeben.
****************************************************************************/
uint8_t TWI_SLA_Received(void)
{
	if (slave_last_state == TRANSMIT)
   {
      // Nach dem Abholen von Daten (Transmit) ist der Slave deaktiviert worden
      // daher den Slave hier neu starten
      TWI_SLA_Start_Transceiver();
      slave_last_state = RECEIVE;
      return(FALSE);
   }
	// wenn head == tail, dann sind keine Daten eingegangen
	// wenn der Slave busy ist, dann ist die bertragung noch nicht abgeschlossen
	if ( (sla_head == sla_tail) || TWI_SLA_Busy() ) return(FALSE);
	else return TRUE;

}


/****************************************************************************
This function is the Interrupt Service Routine (ISR), and called when the TWI interrupt is triggered;
that is whenever a TWI event has occurred. This function should not be called directly from the main
application.
****************************************************************************/

ISR(TWI_vect)
{
switch (TWSR)
	{
// --- SLAVE ------------------------------------------------------------------
// --- TWI Slave Transmitter status codes -------------------------------------

	case TWI_STX_ADR_ACK_M_ARB_LOST: 				// Arbitration lost in SLA+R/W as Master;
																// own SLA+R has been received; ACK has been returned
   case TWI_STX_ADR_ACK:            				// Own SLA+R has been received; ACK has been returned
      slave_last_state = TRANSMIT;
      sla_tail = sla_head;                		// Set buffer pointer to first data location

   case TWI_STX_DATA_ACK:           				// Data byte in TWDR has been transmitted; ACK has been received
      TWI_Mode = (1<<SLAVE | 1<<ACTIV );
																// Es werden solange Daten gesendet, bis der Master mit NACK beendet
      TWDR = sla_buf[sla_head++];
																// der Pointer wird nicht mehr incrementiert, wenn TWI_msgSize erreicht ist
      TWCR =(1<<TWEN)|                         	// TWI Interface enabled
            (1<<TWIE)|(1<<TWINT)|              	// Enable TWI Interupt and clear the flag to send byte
            (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   	// Send ACK
            (0<<TWWC);                         	//
		break;

// case TWI_STX_DATA_ACK_LAST_BYTE: 				// Last data byte in TWDR has been transmitted (TWEA = 0);
																// ACK has been received
   case TWI_STX_DATA_NACK:          				// Data byte in TWDR has been transmitted; NACK has been received.
																// I.e. this could be the end of the transmission.

																// Im Slave-Transmitter Modus beendet der Master mit einem NACK,
																// definiert somit selbst das Ende
																// ein Fehler liegt nur vor, wenn der Master mehr Bytes anfordert,
																// als bereitstehen - das kann der Slave dem Master aber nicht mitteilen ..
		TWI_statusReg.LastTransOK = TRUE;      	// Set status bits to completed successfully.
      TWI_Mode = (1<<SLAVE);
                                                // Put TWI Transceiver in passive mode.
      TWCR = (1<<TWEN)|                         // Enable TWI-interface and release TWI pins
             (0<<TWIE)|(1<<TWINT)|              // Disable Interupt, release SCL
             (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Do not acknowledge on any new requests.
             (0<<TWWC);                         //
		break;


// --- TWI Slave Receiver status codes	----------------------------------------

   case TWI_SRX_GEN_ACK_M_ARB_LOST: 				// Arbitration lost in SLA+R/W as Master;
																// General call address has been received; ACK has been returned
   case TWI_SRX_GEN_ACK:            				// General call address has been received; ACK has been returned
      TWI_statusReg.GenAddressCall = TRUE;

																// ###############################################
   case TWI_SRX_ADR_ACK_M_ARB_LOST: 				// Arbitration lost in SLA+R/W as Master;
																// own SLA+W has been received; ACK has been returned
                                                // Dont need to clear TWI_statusRegister.generalAddressCall
																// due to that it is the default state.

   case TWI_SRX_ADR_ACK: 								// Own SLA+W has been received ACK has been returned
      slave_last_state = RECEIVE;
      sla_tail = sla_head;
      sla_buf[sla_head++] = TWAR;   				// fr durchgngigen Standard: Byte0 = TWI-Adresse
//		TWI_SLA_bufPtr = 1;								// Set buffer pointer/counter to first data location
		TWI_Mode = (1<<SLAVE | 1<<ACTIV );
                                                // Reset the TWI Interupt to wait for a new event.
      TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag to send byte
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Expect ACK on this transmission
             (0<<TWWC);                         //
		break;

   case TWI_SRX_ADR_DATA_ACK:       				// Previously addressed with own SLA+W; data has been received;
																// ACK has been returned
   case TWI_SRX_GEN_DATA_ACK:       				// Previously addressed with general call;
																// data has been received; ACK has been returned
		sla_buf[sla_head++] = TWDR;
		TWI_statusReg.LastTransOK = TRUE;         // Set flag transmission successfull.
                                                // Reset the TWI Interupt to wait for a new event.
		TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag to send byte
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Send ACK after next reception
             (0<<TWWC);                         //
      break;


   case TWI_SRX_STOP_RESTART:       				// A STOP condition or repeated START condition has been received
																// while still addressed as Slave
                                                // Put TWI Transceiver in passive mode.
      TWI_Mode = (1<<SLAVE);
      TWCR = (1<<TWEN)|                         // Enable TWI-interface and release TWI pins
             (0<<TWIE)|(1<<TWINT)|              // Disable Interupt ####
             (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Do not acknowledge on any new requests.
             (0<<TWWC);                         //

		break;
																// nicht behandelte Slave Receive Modi
// case TWI_SRX_ADR_DATA_NACK:      				// Previously addressed with own SLA+W; data has been received;
																// NOT ACK has been returned
// case TWI_SRX_GEN_DATA_NACK:      				// Previously addressed with general call;
																// data has been received; NOT ACK has been returned




// --- MASTER -----------------------------------------------------------------
// ----------------------------------------------------------------------------

// --- General TWI Master status codes	----------------------------------------
	case TWI_START:              						// START has been transmitted
   case TWI_REP_START:          						// Repeated START has been transmitted
      //TWI_bufPtr = twi_tail;           			// Set buffer pointer to the TWI Address location

// --- TWI Master Transmitter status codes ------------------------------------
   case TWI_MTX_ADR_ACK:        						// SLA+W has been transmitted and ACK received
   case TWI_MTX_DATA_ACK:       						// Data byte has been transmitted and ACK received

      if (!(twi_head == twi_tail))              // Daten aus twi_buf schreiben
			{
			TWDR = twi_buf[twi_tail++];      		// Adresse bzw. Daten nach TWDR schreiben
			TWI_Mode = (1<<MASTER | 1<<ACTIV );    // Im Master-Mode
			TWCR =(1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to send byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|	// Send NACK after reception
               (0<<TWWC);
			}

      else                                      // Send STOP after last byte
      {
         TWI_statusReg.LastTransOK = TRUE;      // Set status bits to completed successfully.
			TWI_Mode = (1<<MASTER);                // Der Master ist nicht mehr aktiv

         if (repeated_start)
         {
            TWCR =(1<<TWEN)|                       // TWI Interface enabled
                  (0<<TWIE)|(0<<TWINT)|            // Disable TWI Interrupt and do not clear the flag
                  (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Initiate no STOP condition.
                  (0<<TWWC);
            repeated_start = FALSE;                // nur 1x ausfhren
         }

         else
         {
            TWCR =(1<<TWEN)|                       // TWI Interface enabled
                  (0<<TWIE)|(1<<TWINT)|            // Disable TWI Interrupt and clear the flag
                  (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)| // Initiate a STOP condition.
                  (0<<TWWC);
         }
      }

		break;

// --- TWI Master Receiver status codes ---------------------------------------
   case TWI_MRX_DATA_ACK:                       // Data byte has been received and ACK transmitted
      twi_buf[twi_tail++] = TWDR;     				// gelesene Daten in den Puffer schreiben

   case TWI_MRX_ADR_ACK:                        // SLA+R has been transmitted and ACK received
		TWI_Mode = (1<<MASTER | 1<<ACTIV );
      if (!(twi_head == twi_tail))              // Detect the last byte to NACK it.
			{
			TWCR =(1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
               (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send ACK after reception
               (0<<TWWC);
			}
      else                                      // Send NACK after next reception
			{
			TWCR =(1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send NACK after reception
               (0<<TWWC);
			}
		break;

   case TWI_MRX_DATA_NACK:                      // Data byte has been received and NACK transmitted
      twi_buf[twi_tail++] = TWDR;
      TWI_statusReg.LastTransOK = TRUE;         // Set status bits to completed successfully.
		TWI_Mode = (1<<MASTER);                   // der Master ist nicht mehr aktiv

		TWCR =(1<<TWEN)|                         	// TWI Interface enabled
            (0<<TWIE)|(1<<TWINT)|              	// Disable TWI Interrupt and clear the flag
            (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   	// Initiate a STOP condition.
            (0<<TWWC);
		break;

// --- General TWI Master status codes ----------------------------------------
   case TWI_ARB_LOST:                           // Arbitration lost
		TWI_Mode = (1<<MASTER | 1<<ACTIV );
		TWCR =(1<<TWEN)|                         	// TWI Interface enabled
            (1<<TWIE)|(1<<TWINT)|              	// Enable TWI Interupt and clear the flag
            (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|   	// Initiate a (RE)START condition.
            (0<<TWWC);
      break;

// case TWI_SRX_ADR_DATA_NACK:      				// Previously addressed with own SLA+W; data has been received;
																// NOT ACK has been returned
// case TWI_SRX_GEN_DATA_NACK:      				// Previously addressed with general call;
																// data has been received; NOT ACK has been returned
// case TWI_STX_DATA_ACK_LAST_BYTE: 				// Last data byte in TWDR has been transmitted (TWEA = 0);
																// ACK has been received
// case TWI_BUS_ERROR:         						// Bus error due to an illegal START or STOP condition
// case TWI_MTX_ADR_NACK:                      	// SLA+W has been transmitted and NACK received
// case TWI_MRX_ADR_NACK:                      	// SLA+R has been transmitted and NACK received
// case TWI_MTX_DATA_NACK:                     	// Data byte has been transmitted and NACK received
// case TWI_NO_STATE;                         	// No relevant state information available; TWINT = 0
// case TWI_BUS_ERROR:                        	// Bus error due to an illegal START or STOP condition

   default:
      TWI_state = TWSR;                         // TWSR sichern, wenn Fehler vorliegen
      TWI_Mode = (1<<MASTER);
                                                // Reset TWI Interface
      TWCR =(1<<TWEN)|                         	// Enable TWI-interface and release TWI pins
            (0<<TWIE)|(1<<TWINT)|              	// Disable Interupt   - nderung: clear Interrupt Flag
            (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   	// No Signal requests - nderung: Stop Condition senden
            (0<<TWWC);
				break;
	}
}

// --- [ eof ] ----------------------------------------------------------------
