// --- [ pintools.c ] ----------------------------------------------------------

/*******************************************************************************

Enthlt die Routinen fr die Einstellung der Port-Register:

- DDR-Register  -> Pins als Aus- oder Eingang setzen
- PORT-Register -> Ausgang eines Pins auf high oder low (bwz. Pullup de-/aktivieren)
- PIN-Register  -> an einem Pin anliegenden Pegel (high/low) auslesen

*******************************************************************************/

#include <avr/io.h>
#include <stdlib.h>
#include <stdio.h>
#include <avr/interrupt.h>

#include "global.h"
#include "rs232.h"

// ----------------------------------------------------------------------------
// Setzt das DDR-Register eines Ports und schaltet die Pins
// als Eingangs- oder Ausgangspin
// bergeben wird der Port, die PinMaske (die zu verndernden Pins)
// und die zugehrige Bitmaske
// Parameter: Port [1 .. 3], Pin_Maske, Status_Maske
void port_mode(uint8_t ptr)
{
uint8_t i;

   if (rs_buf[ptr] == 1)                     // PORTB
   {
      i = DDRB;                              // aktuellen Status einlesen
      rs_buf[ptr + 1] &= PORT_B_MASK;        // fest belegte Pins entfernen
      i &= ~rs_buf[(uint8_t)(ptr + 1)];      // alle zu ndernden Pins lschen
      i |= (rs_buf[(uint8_t)(ptr + 2)] & rs_buf[(uint8_t)(ptr + 1)]);// und den neuen Status einkopieren
      DDRB = i;
   }

   else if (rs_buf[ptr] == 2)                // PORTC
   {
      i = DDRC;                              // aktuellen Status einlesen
      rs_buf[ptr + 1] &= PORT_C_MASK;        // fest belegte Pins entfernen
      i &= ~rs_buf[(uint8_t)(ptr + 1)];      // alle zu ndernden Pins lschen
      i |= (rs_buf[(uint8_t)(ptr + 2)] & rs_buf[(uint8_t)(ptr + 1)]);// und den neuen Status einkopieren
      DDRC = i;
   }

   else if (rs_buf[ptr] == 3)                // PORTD
   {
      i = DDRD;                              // aktuellen Status einlesen
      rs_buf[ptr + 1] &= PORT_D_MASK;        // fest belegte Pins entfernen
      i &= ~rs_buf[(uint8_t)(ptr + 1)];      // alle zu ndernden Pins lschen
      i |= (rs_buf[(uint8_t)(ptr + 2)] & rs_buf[(uint8_t)(ptr + 1)]);// und den neuen Status einkopieren
      DDRD = i;
   }

}

// ----------------------------------------------------------------------------
// Liest alle Register eines Port (DDRx, PORTx, PINx)aus und liefert die Werte zurck
// Parameter: Port [1 .. 3], Pin_Maske, Status_Maske
void port_read(uint8_t ptr)
{
   if (rs_buf[ptr] == 1)                   // PORTB
   {
      msg[0] = PORTREAD;
      msg[1] = 4;                         // Anzahl der DatenBytes
      msg[2] = 0x01;                      // PORTB
      msg[3] = DDRB  & PORT_B_MASK;       // PB.6/7 = XTAL
      msg[4] = PORTB & PORT_B_MASK;       // PB.6/7 = XTAL
      msg[5] = PINB  & PORT_B_MASK;       // PB.6/7 = XTAL
      put_msg();
   }

   else if (rs_buf[ptr] == 2)              // PORTC
   {
      msg[0] = PORTREAD;
      msg[1] = 4;                         // Anzahl der DatenBytes
      msg[2] = 0x02;                      // PORTC
      msg[3] = DDRC  & PORT_C_MASK;       // PC.4/5 = SDA/SCL
      msg[4] = PORTC & PORT_C_MASK;       // PC.6 = Reset
      msg[5] = PINC  & PORT_C_MASK;       // PC.7 fehlt
      put_msg();
    }

    else if (rs_buf[ptr] == 3)             // PORTD
    {
      msg[0] = PORTREAD;
      msg[1] = 4;                         // Anzahl der DatenBytes
      msg[2] = 0x03;                      // PORTD
      msg[3] = DDRD  & PORT_D_MASK;       // PD.0/1 = RXD, TXD
      msg[4] = PORTD & PORT_D_MASK;       //
      msg[5] = PIND  & PORT_D_MASK;       //
      put_msg();                          // Parameter: Port [1 .. 3], Pin_Maske);
    }
}


// ----------------------------------------------------------------------------
// Setzt einen oder mehrere Pins (PORT-Register) eines Ports
// Parameter: Port [1 .. 3], Pin_Maske, Status_Maske
void port_write(uint8_t ptr)
{
   uint8_t  j;

   if (rs_buf[ptr] == 1)                     // PORTB
   {
      j = PORTB;                             // aktuellen Portstatus (Ausgnge / Pullups)
      j &= ~rs_buf[(uint8_t)(ptr + 1)];      // alle zu ndernden Pins lschen
      j |= (rs_buf[(uint8_t)(ptr + 2)]);     // und den gewnschten Status addieren
      PORTB = j;                             // Ausgnge / Pullups setzen
   }

   else if (rs_buf[ptr] == 2)                // PORTC
   {
      j = PORTC;                             // aktuellen Portstatus (Ausgnge / Pullups)
      j &= ~rs_buf[(uint8_t)(ptr + 1)];      // alle zu ndernden Pins lschen
      j |= (rs_buf[(uint8_t)(ptr + 2)]);     // und den gewnschten Status
      PORTC = j;                             // Ausgnge / Pullups setzen
   }

   if (rs_buf[ptr] == 3)                     // PORTD
   {
      j = PORTD;                             // aktuellen Portstatus (Ausgnge / Pullups)
      j &= ~rs_buf[(uint8_t)(ptr + 1)];      // alle zu ndernden Pins lschen
      j |= (rs_buf[(uint8_t)(ptr + 2)]);     // und den gewnschten Status addieren
      PORTD = j;                             // Ausgnge / Pullups setzen
   }
}


// ----------------------------------------------------------------------------
// Schaltet einen oder mehrere Pins eines Ports um (toggle)
// Parameter: Port [1 .. 3], Pin_Maske
void port_tog(uint8_t ptr)
{
   if (rs_buf[ptr] == 1)                   // PORTB
   {
      PINB = (rs_buf[(uint8_t)(ptr + 1)] & PORT_B_MASK);
   }

   else  if (rs_buf[ptr] == 2)              // PORTC
   {
      PINC = (rs_buf[(uint8_t)(ptr + 1)] & PORT_C_MASK);
   }

   else  if (rs_buf[ptr] == 3)              // PORTD
   {
      PIND = (rs_buf[(uint8_t)(ptr + 1)] & PORT_D_MASK);
   }
}


// ----------------------------------------------------------------------------
// Konfiguriert einen Interrupt:
// i -> Interrupt [INT0, INT1, PCINT0, PCINT1, PCINT2]
// m ->  fr INTx = sense: [LOWLEVEL, ANYCHANGE, RISINGEDGE, FALLINGEDGE]
//   ->  fr PCINTx = Pinmaske (Pins, an denen ein PinChange einen Interrupt auslsen soll)

void int_set(uint8_t tail)
{
   uint8_t i, m;
   i = rs_buf[tail];
   m = rs_buf[(uint8_t) (tail + 1)];            // berlauf ber 255 bringt Probleme !

   if (i == 0x00)                               // INT0 wird gesetzt
   {
      if (m <= 3)                               // Interrupt setzen, wenn Sense <= 3
      {
         EICRA &= ~(0x03);                      // Sense Control zurcksetzen
         EICRA |= m;                            // neuen Wert setzen
         EIMSK |= (1<<INT0);                    // Flag setzen
      }
      else
      {
         EIMSK &= ~(1<<INT0);                   // Interrupt deaktivieren, wenn Sense > 3
      }
   }

   else if (i == 0x01)                          // INT1 wird gesetzt
   {
      if (m <= 3)                               // Interrupt setzen, wenn Sense <= 3
      {
         EICRA &= ~(0x0C);                      // alten Wert lschen
         EICRA |= (m << 2);                     // Sense Control fr INT1 um 2 Bit nach links schieben
         EIMSK |= (1<<INT1);
      }

      else EIMSK &= ~(1<<INT1);                 // Interrupt deaktivieren, wenn Sense > 3
   }

   else if (i == 0x10)                          // PCINT0 wird gesetzt
   {
      if (m == 0 )                              // wenn Maske == 0, Interrupt deaktivieren
      {
         PCICR &= ~(1<<PCIE0);                  // Interrupt disablen
         PCMSK0 = 0;                            // alle Pins deaktivieren
      }
      else
      {
         PCMSK0 = (m & PORT_B_MASK);
         PCICR |= (1<<PCIE0);
      }
   }

   else if (i == 0x11)                          // PCINT1 wird gesetzt
   {
      if (m == 0 )                              // wenn Maske == 0, Interrupt deaktivieren
      {
         PCICR &= ~(1<<PCIE1);                  // Interrupt disablen
         PCMSK1 = 0;                            // alle Pins deaktivieren
      }
      else
      {
         PCMSK1 = (m & PORT_C_MASK);
         PCICR |= (1<<PCIE1);
      }
   }

   else if (i == 0x12)                          // PCINT2 wird gesetzt
   {
      if (m == 0 )                              // wenn Maske == 0, Interrupt deaktivieren
      {
         PCICR &= ~(1<<PCIE2);                  // Interrupt disablen
         PCMSK2 = 0;                            // alle Pins deaktivieren
      }
      else
      {
         PCMSK2 = (m & PORT_D_MASK);
         PCICR |= (1<<PCIE2);
      }
   }

   else                                         // falscher Interruptpin
   {
      msg[0] = INT_ERR;
      msg[1] = 1;                               // Lnge des Datensatzes
      msg[2] = i;                               // falschen Interrupt zurckgeben
      put_msg();
   }
}


// --- [ eof ] ------------------------------------------------------------------
